
/*
 * Prototypes.
 */

#include "kernel.h"
#include "swis.h"

static int swilist Params((struct b_lelem *bp, _kernel_swi_regs *regs, int size));

/*
 * Swi(swi,r)
 */
FncDcl(Swi,2)
   {
    unsigned int swi_num;
    word nslots;
    int i;

    struct b_list *hp;
    struct b_lelem *bp;

    _kernel_swi_regs regs;

/*
 * Get the SWI number from Arg1.
 */
    if (Qual(Arg1))
    {
	regs.r[1] = (int)StrLoc(Arg1);

	if (_kernel_swi(OS_SWINumberFromString,&regs,&regs))
	    RunErr(205, NULL);

	swi_num = regs.r[0];
    }
    else switch(Type(Arg1))
    {
	case T_Integer:
	    swi_num = (int)IntVal(Arg1);
	    break;
	default:
	    RunErr(123,&Arg1);
    }

/*
 * Make sure that Arg2 is a list, with between 0 and 10 elements, and all
 * are ints.
 */
    if (ChkNull(Arg2))
    {
	for (i = 0; i < 10; ++i)
	    regs.r[i] = 0;
    }
    else
    {
	if (Type(Arg2) != T_List) {
	    RunErr(108, &Arg2);
	}

	hp = (struct b_list *) BlkLoc(Arg2);
	if (hp->size < 0 || hp->size > 10) {
	    RunErr(205, &Arg1);
	}

	bp = (struct b_lelem *) hp->listhead;

	if (swilist(bp, &regs, hp->size) == Failure)
	    Fail;
    }

    if (_kernel_swi(swi_num,&regs,&regs) != NULL)
       Fail;

/*
 * Return the values.
 */
    nslots = 10;
    if (blkreq((word)sizeof(struct b_list) + sizeof(struct b_lelem) +
	nslots * sizeof(struct descrip)) == Error)
    {
	RunErr(0, NULL);
    }

    hp = alclist((word)10);
    bp = alclstb(nslots,(word)0,(word)10);
    hp->listhead = hp->listtail = (union block *) bp;

/* returns [r0,r1,r2,r3,r4,r5,r6,r7,r8,r9] */

    MakeInt((uword)regs.r[0],&(bp->lslots[0]));
    MakeInt((uword)regs.r[1],&(bp->lslots[1]));
    MakeInt((uword)regs.r[2],&(bp->lslots[2]));
    MakeInt((uword)regs.r[3],&(bp->lslots[3]));
    MakeInt((uword)regs.r[4],&(bp->lslots[4]));
    MakeInt((uword)regs.r[5],&(bp->lslots[5]));
    MakeInt((uword)regs.r[6],&(bp->lslots[6]));
    MakeInt((uword)regs.r[7],&(bp->lslots[7]));
    MakeInt((uword)regs.r[8],&(bp->lslots[8]));
    MakeInt((uword)regs.r[9],&(bp->lslots[9]));

    ArgType(0) = D_List;
    Arg(0).vword.bptr = (union block *) hp;
    Return;
   }

/*
 * peek(addr,len)
 */

FncDcl(Peek,2)
{
   if (defshort(&Arg2,1) == Error) {
       RunErr(0, NULL);
   }

   switch (Type(Arg1)) {
	case T_Integer:
		StrLoc(Arg0) = (char *)(Arg1.vword.integr);
	    break;
	default:
	    RunErr(101,&Arg1);
   }

   switch (Type(Arg2)) {
	case T_Integer:
		StrLen(Arg0) = Arg2.vword.integr;
	    break;
	default:
	    RunErr(101, &Arg2);
   }

   Return;
}

/*
 * poke(addr,s)
 */

FncDcl(Poke,2)
{
   register char *s1,*s2;
   register word l;
   char *addr;

   switch (Type(Arg1)) {
	case T_Integer:
		addr = (char *)(Arg1.vword.integr);
	    break;
	default:
	    RunErr(101,&Arg1);
   }

   if (!Qual(Arg2)) {
      RunErr(103, &Arg2);
   }

   l = StrLen(Arg2);
   s1 = StrLoc(Arg2);
   s2 = addr;

   memcopy(s2,s1,l);	 /* Copy... */

   Return;
}

/*
 * Getspace(i)
 */

FncDcl(GetSpace,1)
{
      char *addr;
      uword u;
      int size;

      if (Arg1.dword != D_Integer) {	/* should convert & check for small */
	   RunErr(101, &Arg1);
      }

      size = (int)Arg1.vword.integr;
      addr = (char *)calloc(size,sizeof(char));
      if (addr==NULL) Fail;
      u = (uword)addr;
      MakeInt(u,&Arg0);
      Return;
}

/*
 * FreeSpace(a)
 */

FncDcl(FreeSpace,1)
{
   uword u;
   char *addr;

   switch (Type(Arg1)) {
	case T_Integer:
	   u = Arg1.vword.integr;
	   break;
	default:
	    RunErr(101, &Arg1);
   }
   addr = (char *)u;
   free((pointer)addr);
   Return;
}

static int swilist(bp,regs,size)
int size;
_kernel_swi_regs *regs;
struct b_lelem *bp;
{
    int i;
    int count;

    i = 0;
    for (count = 0; count < size; ++count)
    {
	int j;
	if (++i > bp->nused) {
	    i = 1;
	    bp = (struct b_lelem *) bp->listnext;
	}

	j = bp->first + i - 1;	/* Get slot index */

	if (j >= bp->nslots)
	    j -= bp->nslots;

	if (Qual(bp->lslots[j]))
	    regs->r[count] = (int)StrLoc(bp->lslots[j]);

	else switch(Type(bp->lslots[j]))
	{
	    case T_Integer:
		regs->r[count] = (int)IntVal(bp->lslots[j]);
		break;
	    default:
		RunErr(123,&bp->lslots[j]);
	}
   }

   for (i = count; i < 10; ++i)
      regs->r[i] = 0;

   return 0;
}

