/*
 * File: fsys.c
 *  Contents: close, exit, getenv, open, read, reads, [save], stop, system,
 *   write, writes
 */

#include "rt.h"

/*
 * The following code is operating-system dependent. Include system-dependent
 *  files.
 */

#if PORT
   /* nothing to do */
#endif					/* PORT */

#if MACINTOSH
#if MPW
#include <IOCtl.h>
#define isatty(fd) (!ioctl((fd), FIOINTERACTIVE))
#endif					/* MPW */
#endif					/* MACINTOSH */

#if AMIGA || ATARI_ST || MSDOS || UNIX || VMS
   /* nothing to do */
#endif					/* AMIGA || ATARI_ST || MSDOS || ... */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */



extern char *alcstr();


/*
 * close(f) - close file f.
 */

FncDcl(close,1)
   {
   register int i;

   /*
    * Arg1 must be a file.
    */
   if (Arg1.dword != D_File) {
      runerr(105, &Arg1);
      Fail;
      }

   /*
    * If Arg1 has a buffer associated with it, "free" it.
    */
   for (i = 0; i < numbufs; i++) {
      if (bufused[i] == BlkLoc(Arg1)->file.fd) {
         bufused[i] = NULL;
         break;
         }
      }

   /*
    * Close Arg1, using fclose or pclose as appropriate.
    */

/*
 * The following code is operating-system dependent. Close pipe if
 *  pipes are supported.
 */

#if PORT
   /* not supported */
#endif					/* PORT */

#if AMIGA || ATARI_ST || MACINTOSH || MSDOS
   /* not supported */
#endif					/* AMIGA || ATARI_ST || MACINTOSH ... */

#if UNIX || VMS
   if (BlkLoc(Arg1)->file.status & Fs_Pipe) {
      BlkLoc(Arg1)->file.status = 0;
      MkInt((long)((pclose(BlkLoc(Arg1)->file.fd) >> 8) & 0377), &Arg0);
      Return;
      }
   else
#endif					/* UNIX || VMS */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */

      fclose(BlkLoc(Arg1)->file.fd);
   BlkLoc(Arg1)->file.status = 0;

   /*
    * Return the closed file.
    */
   Arg0 = Arg1;
   Return;
   }

#ifdef CanDump
/*
 * save(s) - save the run-time system in file s
 */

FncDcl(save,1)
   {
   char sbuf[MaxCvtLen];
   int f, fsz;

   dumped = 1;

   /* if (ChkNull(Arg1)) { abort(); } */

   /*
    * Make a string out of Arg1
    */
   switch (cvstr(&Arg1, sbuf)) {

      case Cvt:
      case NoCvt:
 	 qtos(&Arg1, sbuf);
	 break;

      default:
         runerr(103, &Arg1);
         Fail;
      }


   /*
    * Open the file for the executable image.
    */
   f = creat(sbuf, 0777);
   if (f == -1)
      Fail;
   fsz = wrtexec(f);
   /*
    * It happens that most wrtexecs don't check the system call return
    *  codes and thus they'll never return -1.  Nonetheless...
    */
   if (fsz == -1)
      Fail;
   /*
    * Return the size of the data space.
    */
   MkInt(fsz, &Arg0);
   Return;
   }

#endif					/* CanDump */

/*
 * exit(status) - exit process with specified status, defaults to 0.
 */

FncDcl(exit,1)
   {
   if (defshort(&Arg1, NormalExit) == Failure)
      Fail;
   c_exit((int)IntVal(Arg1));
   }

/*
 * getenv(s) - return contents of environment variable s
 */

FncDcl(getenv,1)
   {
#ifdef EnvVars
   register char *p;
   register word len;
   char sbuf[256];
   extern char *getenv();
   extern char *alcstr();

   if (!Qual(Arg1)) {			/* check legality of argument */
      runerr(103, &Arg1);
      Fail;
      }
   if (StrLen(Arg1) <= 0 || StrLen(Arg1) >= MaxCvtLen) {
      runerr(401, &Arg1);
      Fail;
      }
   qtos(&Arg1, sbuf);			/* convert argument to C-style string */

   if ((p = getenv(sbuf)) != NULL) {	/* get environment variable */
      len = strlen(p);
      strreq(len);
      StrLen(Arg0) = len;
      StrLoc(Arg0) = alcstr(p, len);
      Return;
      }
   else 				/* fail if not in environment */
      Fail;
#else					/* EnvVars */
   runerr(121, NULL);
   Fail;
#endif					/* EnvVars */
   }

/*
 * open(s1,s2) - open file s1 with specification s2.
 */
FncDcl(open,2)
   {
   register word slen;
   register int i;
   register char *s;
   int status;

/*
 * The following code is operating-system dependent. Make declarations as
 *  needed for opening files.
 */

#if PORT
   char mode[3];
   extern FILE *fopen();
#endif					/* PORT */

#if AMIGA
   char mode[3];
   extern FILE *fopen();
#endif					/* AMIGA */

#if ATARI_ST || MSDOS
   char mode[4];
   char untranslated;
   extern FILE *fopen();
#endif					/* ATARI_ST || MSDOS */

#if MACINTOSH || UNIX || VMS
   char mode[3];
   extern FILE *fopen(), *popen();
#endif					/* MACINTOSH || UNIX || VMS */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */

   char sbuf1[MaxCvtLen], sbuf2[MaxCvtLen];
   struct descrip cstring;
   char *openstring = sbuf1;
   FILE *f;
   extern struct b_file *alcfile();
   extern struct descrip tended[];

   /*
    * Arg1 must be a string and a C string copy of it is also needed.
    *  Make it a string if it is not one; make a C string if Arg1 is
    *  a string.
    */
   switch (cvstr(&Arg1, sbuf1)) {

      case Cvt:
         strreq(StrLen(Arg1));
         StrLoc(Arg1) = alcstr(StrLoc(Arg1), StrLen(Arg1));
         break;

      case NoCvt:
 	 if (StrLen(Arg1) >= MaxCvtLen) {
 	    /*
 	     * Make a C string from argument 1, which is type string.
 	     * If the string is longer than MaxCvtLen, the C string is
 	     * created by creating a new string is string space with the
 	     * terminator '\0' concatenated to the end.  Otherwise the C
 	     * string is built in sbuf1.  If we build a string in string
 	     * space, its qualifier must be tended since a garbage
 	     * collection might occur before we use it.
 	     */
 	    StrLen(cstring) = StrLen(Arg1) + 1;
 	    strreq(StrLen(cstring));
 	    StrLoc(cstring) = StrLoc(Arg1) + StrLen(Arg1) == strfree ?
 		StrLoc(Arg1) : alcstr(StrLoc(Arg1),StrLen(Arg1));
 	    alcstr("",(word)1);
 	    tended[1] = cstring;
 	    ntended = 1;
 	    openstring = StrLoc(cstring);
 	    }
 	 else
 	    qtos(&Arg1, sbuf1);
         break;

      default:
         runerr(103, &Arg1);
         Fail;
      }
   /*
    * s2 defaults to "r".
    */
   if (defstr(&Arg2, sbuf2, &letr) == Failure)
      Fail;

   blkreq((word)sizeof(struct b_file));
   status = 0;

/*
 * The following code is operating-system dependent. Provide declaration
 *  for untranslated line-termination mode, if supported.
 */

#if PORT
   /* nothing to do */
#endif					/* PORT */

#if AMIGA
   /* translated mode could be supported, but is not now */
#endif					/* AMIGA */

#if ATARI_ST || MSDOS
   untranslated = 0;
#endif					/* ATARI_ST || MSDOS */

#if MACINTOSH || UNIX || VMS
   /* nothing to do */
#endif					/* MACINTOSH || UNIX || VMS */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */

   /*
    * Scan Arg2, setting appropriate bits in status.  Produce a run-time error
    *  if an unknown character is encountered.
    */
   s = StrLoc(Arg2);
   slen = StrLen(Arg2);
   for (i = 0; i < slen; i++) {
      switch (*s++) {
         case 'a': case 'A':
            status |= Fs_Write|Fs_Append;
            continue;
         case 'b': case 'B':
            status |= Fs_Read|Fs_Write;
            continue;
         case 'c': case 'C':
            status |= Fs_Create|Fs_Write;
            continue;

/*
 * The following code is operating-system dependent. Handle untranslated
 *  line-terminator mode and pipes, if supported.
 */

#if PORT
   /* not supported */
#endif					/* PORT */

#if AMIGA || MACINTOSH
   /* not supported */
#endif					/* AMIGA || MACINTOSH */

#if ATARI_ST || MSDOS
         case 't': case 'T':   /* Add a translated (default) */
            untranslated = 0;
            continue;
         case 'u': case 'U':   /* Add a translated (default) */
            untranslated = 1;
            continue;
#endif					/* ATARI_ST || MSDOS */

#if UNIX || VMS
         case 'p': case 'P':
            status |= Fs_Pipe;
            continue;
#endif					/* UNIX || VMS */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */
         case 'r': case 'R':
            status |= Fs_Read;
            continue;
         case 'w': case 'W':
            status |= Fs_Write;
            continue;
         default:
            runerr(209, &Arg2);
            Fail;
         }
      }

   /*
    * Construct a mode field for fopen/popen.
    */
   mode[0] = '\0';
   mode[1] = '\0';
   mode[2] = '\0';

/*
 * The following code is operating-system dependent. Set mode[3] if used.
 */

#if PORT
   /* nothing to do */
#endif					/* PORT */

#if ATARI_ST || MSDOS
   mode[3] = '\0';
#endif					/* ATARI_ST || MSDOS */

#if AMIGA || MACINTOSH || UNIX || VMS
   /* nothing to do */
#endif					/* AMIGA || MACINTOSH || UNIX || VMS */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */

   if ((status & (Fs_Read|Fs_Write)) == 0)   /* default: read only */
      status |= Fs_Read;
   if (status & Fs_Create)
      mode[0] = 'w';
   else if (status & Fs_Append)
      mode[0] = 'a';
   else if (status & Fs_Read)
      mode[0] = 'r';
   else
      mode[0] = 'w';

/*
 * The following code is operating-system dependent. Handle open modes.
 */

#if PORT
   if ((status & (Fs_Read|Fs_Write)) == (Fs_Read|Fs_Write))
      mode[1] = '+';
#endif					/* PORT */

#if ATARI_ST
   if ((status & (Fs_Read|Fs_Write)) == (Fs_Read|Fs_Write)) {
      mode[1] = '+';
      mode[2] = untranslated ? 'b' : 'a';
      }
   else mode[1] = untranslated ? 'b' : 'a';
#endif					/* ATARI_ST */

#if MSDOS
   if ((status & (Fs_Read|Fs_Write)) == (Fs_Read|Fs_Write)) {
      mode[1] = '+';
      mode[2] = untranslated ? 'b' : 't';
      }
   else mode[1] = untranslated ? 'b' : 't';
#endif					/* MSDOS */

#if AMIGA || MACINTOSH || UNIX || VMS
   if ((status & (Fs_Read|Fs_Write)) == (Fs_Read|Fs_Write))
      mode[1] = '+';
#endif					/* AMIGA || MACINTOSH || UNIX || VMS */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */

   /*
    * Open the file with fopen or popen.
    */

/*
 * The following code is operating-system dependent. Handle opening pipes
 * if supported.
 */

#if PORT
   /* not supported */
#endif					/* PORT */

#if AMIGA || ATARI_ST || MACINTOSH || MSDOS
   /* not supported */
#endif					/* AMIGA || ATARI_ST || MACINTOSH ... */

#if UNIX || VMS
   if (status & Fs_Pipe) {
      if (status != (Fs_Read|Fs_Pipe) && status != (Fs_Write|Fs_Pipe)) {
         runerr(209, &Arg2);
         Fail;
         }
      f = popen(openstring, mode);
      }
   else
#endif					/* UNIX || VMS */

#if VM || MVS
#endif					/* VM || MVS */


/*
 * End of operating-system specific code.
 */

      f = fopen(openstring, mode);
   /*
    * Fail if the file cannot be opened.
    */
   if (f == NULL)
      Fail;
   /*
    * If the file is not a terminal and a buffer is available, assign
    *  it to the file.
    */

/*
 * The following code is operating-system dependent. It handles buffering.
 */

#if PORT
   /* not supported */
#endif					/* PORT */

#if AMIGA
   /* no isatty function so skip this for now */
#endif					/* AMIGA */

#if ATARI_ST
   for (i = 0; i < numbufs; i++)
      if (bufused[i] == NULL)
         break;
   if (i < numbufs) {	      /* Use buffer if any free. */
      setbuf(f, bufs[i]);
      bufused[i] = f;
      }
#endif					/* ATARI_ST */

#if MACINTOSH
#if MPW
   for (i = 0; i < numbufs; i++)
      if (bufused[i] == NULL)
         break;
   if (i < numbufs) {	      /* Use buffer if any free. */
      setbuf(f, bufs[i]);
      bufused[i] = f;
      }
#endif					/* MPW */
#endif					/* MACINTOSH */

#if UNIX || MSDOS
   if (!isatty(fileno(f))) {
      for (i = 0; i < numbufs; i++)
         if (bufused[i] == NULL)
            break;
      if (i < numbufs) {	      /* Use buffer if any free. */
         setbuf(f, bufs[i]);
         bufused[i] = f;
         }
      else
         setbuf(f, NULL);
      }
   else
      setbuf(f, NULL);
#endif					/* UNIX || MSDOS */

#if VMS
   /* not supported */
#endif					/* VMS */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */
   /*
    * Return the resulting file value.
    */
   Arg0.dword = D_File;
   BlkLoc(Arg0) = (union block *) alcfile(f, status, &Arg1);
   Return;
   }

/*
 * read(f) - read line on file f.
 */
FncDcl(read,1)
   {
   register word slen, rlen;
   register char *sp;
   int status;
   static char sbuf[MaxReadStr];
   FILE *f;
   extern char *alcstr();

   /*
    * Default Arg1 to &input.
    */
   if (deffile(&Arg1, &input) == Failure)
      Fail;
   /*
    * Get a pointer to the file and be sure that it is open for reading.
    */
   f = BlkLoc(Arg1)->file.fd;
   status = BlkLoc(Arg1)->file.status;
   if ((status & Fs_Read) == 0) {
      runerr(212, &Arg1);
      Fail;
      }

   /*
    * Use getstr to read a line from the file, failing if getstr
    *  encounters end of file.
    */
   StrLen(Arg0) = 0;
   do {
      if ((slen = getstr(sbuf,MaxReadStr,f)) == -1)
	 Fail;
      /*
       * Allocate the string read and make Arg0 a descriptor for it.
       */
      rlen = slen < 0 ? MaxReadStr : slen;
      strreq(rlen);
      sp = alcstr(sbuf,rlen);
      if (StrLen(Arg0) == 0) StrLoc(Arg0) = sp;
      StrLen(Arg0) += rlen;
      } while (slen < 0);
   Return;
   }

/*
 * reads(f,i) - read i characters on file f.
 */
FncDcl(reads,2)
   {
   register int cnt;
   int status;
   FILE *f;

   /*
    * Arg1 defaults to &input and Arg2 defaults to 1 (character).
    */
   if (deffile(&Arg1, &input) == Failure)
      Fail;
   if (defshort(&Arg2, 1) == Failure)
      Fail;

   /*
    * Get a pointer to the file and be sure that it is open for reading.
    */
   f = BlkLoc(Arg1)->file.fd;
   status = BlkLoc(Arg1)->file.status;
   if ((status & Fs_Read) == 0) {
      runerr(212, &Arg1);
      Fail;
      }

   /*
    * Be sure that a positive number of bytes is to be read.
    */
   if ((cnt = IntVal(Arg2)) <= 0) {
      runerr(205, &Arg2);
      Fail;
      }

   /*
    * Ensure that enough space for the string exists and read it directly
    *  into the string space.  (By reading directly into the string space,
    *  no arbitrary restrictions are placed on the size of the string that
    *  can be read.)  Make Arg0 a descriptor for the string and return it.
    */
   strreq((word)cnt);
   if (strfree + cnt > strend)
      syserr("string allocation botch");
   StrLoc(Arg0) = strfree;

/*
 * The following code is operating-system dependent. Interactive files
 *  may need special handling.
 */

#if PORT
   /* probably nothing to do */
#endif					/* PORT */

#if AMIGA
   if (IsInteractive(fileno(f))) {
      if ((cnt = read(fileno(f),StrLoc(Arg0),cnt)) <= 0)
         Fail;
      }
   else
#endif					/* AMIGA */

#if ATARI_ST || MACINTOSH || MSDOS || UNIX || VMS
   /* nothing to do */
#endif					/* ATARI_ST || MACINTOSH || MSDOS ... */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */
   if ((cnt = fread(StrLoc(Arg0), sizeof(char), cnt, f)) <= 0)
      Fail;
   StrLen(Arg0) = cnt;
   strfree += cnt;
   Return;
   }

/*
 * stop(a,b,...) - write arguments (starting on error output) and stop.
 */

FncDclV(stop)
    {
   register word n;
   char sbuf[MaxCvtLen];
   struct descrip arg;
   FILE *f;

   f = stderr;
   /*
    * Loop through arguments.
    */

   for (n = 1; n <= nargs; n++) {
      arg = Arg(n);
      if (arg.dword == D_File) {
         if (n > 1)
            putc('\n', f);
         if ((BlkLoc(arg)->file.status & Fs_Write) == 0) {
            runerr(213, &arg);
            Fail;
            }
         f = BlkLoc(arg)->file.fd;
         }
      else {
         if (n == 1 && (k_output.status & Fs_Write) == 0) {
            runerr(213, NULL);
            Fail;
            }
         if (ChkNull(arg))
            arg = emptystr;
         if (cvstr(&arg, sbuf) == CvtFail) {
            runerr(109, &arg);
            Fail;
            }
         putstr(f, StrLoc(arg), StrLen(arg));
         }
      }

   putc('\n', f);
   c_exit(ErrorExit);
   }

/*
 * system(s) - execute string s as a system command.
 */

FncDcl(system,1)
   {
   char sbuf1[MaxCvtLen], sbuf2[MaxCvtLen];
   struct descrip cstring;
   char *systemstring = sbuf2;

/*
 * The following code is operating-system dependent. Provide declarations
 *  related to system calls if needed.
 */

#if PORT
   /* nothing to do */
#endif					/* PORT */

#if AMIGA || ATARI_ST || MACINTOSH || MSDOS || UNIX
   /* nothing to do */
#endif					/* AMIGA || ATARI_ST || MACINTOSH ... */

#if VMS
    struct { long size; char *ptr; } descr;
    int status;
#endif					/* VMS */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */

   /*
    * Arg1 must be string.
    */
   if (cvstr(&Arg1, sbuf1) == CvtFail) {
      runerr(103, &Arg1);
      Fail;
      }
   if (StrLen(Arg1) >= MaxCvtLen) {
      /*
       * Make a C string from argument 1, which is type string.
       * If the string is longer than MaxCvtLen, the C string is
       * created by creating a new string is string space with the
       * terminator '\0' concatenated to the end.  Otherwise the C
       * string is built in sbuf2.
       */
      StrLen(cstring) = StrLen(Arg1) + 1;
      strreq(StrLen(cstring));
      StrLoc(cstring) = StrLoc(Arg1) + StrLen(Arg1) == strfree ?
         StrLoc(Arg1) : alcstr(StrLoc(Arg1),StrLen(Arg1));
      alcstr("",(word)1);
      systemstring = StrLoc(cstring);
      }
    else
       qtos(&Arg1, sbuf2);

   /*
    * Pass the C string to the system() function and return the exit code
    *  of the command as the result of system().
    */

/*
 * The following code is operating-system dependent. Perform system call.
 *  Should be runerr(121, NULL) if not supported.
 */

#if PORT
   runerr(121, NULL);
   Fail;
#endif					/* PORT */

#if ATARI_ST
   MkInt(system(systemstring), &Arg0);
#endif					/* ATARI_ST */

#if MACINTOSH
#if MPW
   runerr(121, NULL);
   Fail;
#endif					/* MPW */
#endif					/* MACINTOSH */

#if AMIGA || UNIX || MSDOS
   MkInt((long)((system(systemstring) >> 8) & 0377), &Arg0);
   Return;
#endif					/* AMIGA || UNIX || MSDOS */

#if VM || MVS
#endif					/* VM || MVS */

#if VMS
   descr.ptr = systemstring;
   descr.size = StrLen(Arg1);
   status = 0;
   status = lib$spawn(&descr,0,0,0,0,0, &status,0,0,0);
   MkInt((long)status|ErrorExit, &Arg0);
   Return;
#endif					/* VMS */

/*
 * End of operating-system specific code.
 */
   }

/*
 * write(a,b,...) - write arguments.
 */
FncDclV(write)
   {
   register word n;
   char sbuf[MaxCvtLen];
   struct descrip arg;
   FILE *f;
   extern char *alcstr();

   f = stdout;
   arg = emptystr;

   /*
    * Loop through the arguments.
    */
   for (n = 1; n <= nargs; n++) {
      arg = Arg(n);
      if (arg.dword == D_File)	{	/* Current argument is a file */
         /*
          * If this is not the first argument, output a newline to the current
          *  file and flush it.
          */
         if (n > 1) {
            putc('\n', f);
            fflush(f);
            }
         /*
          * Switch the current file to the file named by the current argument
          *  providing it is a file.  arg is made to be a empty string to
          *  avoid a special case.
          */
         if ((BlkLoc(arg)->file.status & Fs_Write) == 0) {
            runerr(213, &arg);
            Fail;
            }
         f = BlkLoc(arg)->file.fd;
         arg = emptystr;
         }
      else {	/* Current argument is a string */
         /*
          * On first argument, check to be sure that &output is open
          *  for output.
          */
         if (n == 1 && (k_output.status & Fs_Write) == 0) {
            runerr(213, NULL);
            Fail;
            }
         /*
          * Convert the argument to a string, defaulting to a empty string.
          */
         if (ChkNull(arg))
            arg = emptystr;
         if (cvstr(&arg, sbuf) == CvtFail) {
            runerr(109, &arg);
            Fail;
            }
         /*
          * Output the string.
          */
         putstr(f, StrLoc(arg), StrLen(arg));
         }
      }
   /*
    * Append a newline to the file and flush it.
    */
   putc('\n', f);
   fflush(f);

   /*
    * If the beginning of the last string output lies in sbuf,
    *  allocate it as a real string.  Note that some of the string
    *  conversions do not always leave the converted string at the
    *  start of the conversion buffer, hence the range check.
    */
   if (StrLoc(arg) >= sbuf && StrLoc(arg) < sbuf + MaxCvtLen) {
      strreq(StrLen(arg));
      StrLoc(arg) = alcstr(StrLoc(arg), StrLen(arg));
      }
   /*
    * Return the string corresponding to the last argument.
    */
   Arg(0) = arg;
   Return;
   }

/*
 * writes(a,b,...) - write arguments without newline terminator.
 */

FncDclV(writes)
   {
   register word n;
   char sbuf[MaxCvtLen];
   struct descrip arg;
   FILE *f;
   extern char *alcstr();

   f = stdout;
   arg = emptystr;

   /*
    * Loop through the arguments.
    */
   for (n = 1; n <= nargs; n++) {
      arg = Arg(n);
      if (arg.dword == D_File)	{	/* Current argument is a file */
         /*
          * Switch the current file to the file named by the current argument
          *  providing it is a file.  arg is made to be a empty string to
          *  avoid a special case.
          */
         if ((BlkLoc(arg)->file.status & Fs_Write) == 0) {
            runerr(213, &arg);
            Fail;
            }
         f = BlkLoc(arg)->file.fd;
         arg = emptystr;
         }
      else {	/* Current argument is a string */
         /*
          * On first argument, check to be sure that &output is open
          *  for output.
          */
         if (n == 1 && (k_output.status & Fs_Write) == 0) {
            runerr(213, NULL);
            Fail;
            }
         /*
          * Convert the argument to a string, defaulting to a empty string.
          */
         if (ChkNull(arg))
            arg = emptystr;
         if (cvstr(&arg, sbuf) == CvtFail) {
            runerr(109, &arg);
            Fail;
            }
         /*
          * Output the string and flush the file.
          */
         putstr(f, StrLoc(arg), StrLen(arg));
         fflush(f);
         }
      }
   /*
    * If the beginning of the last string output lies in sbuf,
    *  allocate it as a real string.  Note that some of the string
    *  conversions do not always leave the converted string at the
    *  start of the conversion buffer, hence the range check.
    */
   if (StrLoc(arg) >= sbuf && StrLoc(arg) < sbuf + MaxCvtLen) {
      strreq(StrLen(arg));
      StrLoc(arg) = alcstr(StrLoc(arg), StrLen(arg));
      }
   /*
    * Return the string corresponding to the last argument.
    */
   Arg(0) = arg;
   Return;
   }
