/*
 * Initialization and error routines.
 */

#include "rt.h"
#include "version.h"
#include "gc.h"
#include "header.h"
#include "opdefs.h"
#include <ctype.h>

#ifdef MemMon
#define IconInit MMinit
#define IconTerm MMterm
#else						/* MemMon */
#define IconInit()
#define IconTerm()
#endif						/* MemMon */

/*
 * The following code is operating-system dependent. Include files and
 *  declarations that are are system-dependent.
 */

#if PORT
#include <signal.h>
   /* probably needs something more */
#endif					/* PORT */

#if AMIGA
#include <signal.h>
#include <time.h>
#include <fcntl.h>

int chkbreak;				/* if nonzero, check for ^C */
#endif					/* AMIGA */

#if ATARI_ST
#include "fcntl.h"
struct tms {
   unsigned int tms_utime;		/* user time */
   int tm_year;
   int tm_mon;
   int tm_mday;
   int tm_hour;
   int tm_min;
   int tm_sec;
   };
struct tms *tbuf;

unsigned long _STACK = 10240;
unsigned long _MNEED = 200000;
#endif					/* ATARI_ST */

#if MACINTOSH
#include <signal.h>
#if MPW
#include <Types.h>
#include <Events.h>
#include <FCntl.h>
#include <SANE.h>
#include SysTime
#endif					/* MPW */
#endif					/* MACINTOSH */

#if MSDOS
#include <fcntl.h>
#include <signal.h>
#include <time.h>
#if MICROSOFT
#include <sys/types.h>
#endif					/* MICROSOFT */
#endif					/* MSDOS */

#if UNIX
#include <signal.h>
#include <sys/types.h>
#include <sys/times.h>
#endif					/* UNIX */

#if VM || MVS
#endif					/* VM || MVS */

#if VMS
#include <signal.h>
#include <types.h>
struct tms {
   time_t tms_utime;		/* user time */
   time_t tms_stime;		/* system time */
   time_t tms_cutime;		/* user time, children */
   time_t tms_cstime;		/* system time, children */
   };
#endif					/* VMS */

/*
 * End of operating-system specific code.
 */

#ifdef IconAlloc
#define malloc mem_alloc
#endif					/* IconAlloc */

#ifndef MaxHeader
#define MaxHeader MaxHdr
#endif					/* MaxHeader */

/*
 * A number of important variables follow.
 */

#ifndef NoTraceBack
extern struct b_proc *opblks[];
extern word lastop;
extern struct descrip *xargp;
extern word xnargs;
#endif					/* NoTraceBack */

extern putpos();			/* assignment function for &pos */
extern putran();			/* assignment function for &random */
extern putsub();			/* assignment function for &subject */
extern puttrc();			/* assignment function for &trace */
extern puterr();			/* assignment function for &error */

word *stack;				/* interpreter stack */
int line = 0;				/* source program line number */
int k_level = 0;			/* &level */
int k_errornumber = 0;			/* &errornumber */
char *k_errortext = "";			/* &errortext */
struct descrip k_errorvalue;		/* &errorvalue */
struct descrip k_main;			/* &main */
char *code;				/* interpreter code buffer */
word *records;				/* pointer to record procedure blocks */
word *ftab;				/* pointer to record/field table */
struct descrip *globals, *eglobals;	/* pointer to global variables */
struct descrip *gnames, *egnames;	/* pointer to global variable names */
struct descrip *statics, *estatics;	/* pointer to static variables */
char *strcons;				/* pointer to string constant table */

int numbufs = NumBuf;			/* number of i/o buffers */
char (*bufs)[BUFSIZ];			/* pointer to buffers */
FILE **bufused;				/* pointer to buffer use markers */

word tallybin[16];			/* counters for tallying */
int tallyopt = 0;			/* want tally results output? */

word mstksize = MStackSize;		/* initial size of main stack */
word stksize = StackSize;		/* co-expression stack size */
struct b_coexpr *stklist;		/* base of co-expression block list */

word statsize = MaxStatSize;		/* size of static region */
word statincr = MaxStatSize/4;		/* increment for static region */
char *statbase;				/* start of static space */
char *statend;				/* end of static space */
char *statfree;				/* static space free pointer */

word ssize = MaxStrSpace;		/* initial string space size (bytes) */
char *strbase;				/* start of string space */
char *strend;				/* end of string space */
char *strfree;				/* string space free pointer */
char *currend = NULL;			/* current end of memory region */

word abrsize = MaxAbrSize;		/* initial size of allocated block
						region (bytes) */
char *blkbase;				/* start of block region */
char *blkend;				/* end of allocated blocks */
char *blkfree;				/* block region free pointer */

#ifdef FixedRegions
word qualsize = QualLstSize;		/* size of quallist for fixed regions */
#endif					/* FixedRegions */

uword statneed;				/* stated need for static space */
uword strneed;				/* stated need for string space */
uword blkneed;				/* stated need for block space */

struct descrip **quallist;		/* string qualifier list */
struct descrip **qualfree;		/* qualifier list free pointer */
struct descrip **equallist;		/* end of qualifier list */

int dodump;				/* if nonzero, core dump on error */
int noerrbuf;				/* if nonzero, do not buffer stderr */

struct descrip k_current;		/* current expression stack pointer */
struct descrip maps2;			/* second cached argument of map */
struct descrip maps3;			/* third cached argument of map */

int ntended = 0;			/* number of active tended descrips */
long starttime;				/* start time of job in milliseconds */

#ifdef CanDump
int dumped = 0;				/* non-zero if reloaded from dump */
#endif					/* CanDump */

/*
 * Run-time error numbers and text.
 */
struct errtab {
   int errno;
   char *errmsg;
   } errtab[] = {
   101, "integer expected",
   102, "numeric expected",
   103, "string expected",
   104, "cset expected",
   105, "file expected",
   106, "procedure or integer expected",
   107, "record expected",
   108, "list expected",
   109, "string or file expected",
   110, "string or list expected",
   111, "variable expected",
   112, "invalid type to size operation",
   113, "invalid type to random operation",
   114, "invalid type to subscript operation",
   115, "list or table expected",
   116, "invalid type to element generator",
   117, "missing main procedure",
   118, "co-expression expected",
   119, "set expected",
   120, "cset or set expected",
   121, "function not supported",

   201, "division by zero",
   202, "remaindering by zero",
   203, "integer overflow",
   204, "real overflow, underflow, or division by zero",
   205, "value out of range",
   206, "negative first operand to real exponentiation",
   207, "invalid field name",
   208, "second and third arguments to map of unequal length",
   209, "invalid second argument to open",
   210, "argument to system function too long",
   211, "by clause equal to zero",
   212, "attempt to read file not open for reading",
   213, "attempt to write file not open for writing",
   214, "activation of an active co-expression",

   301, "interpreter stack overflow",
   302, "C stack overflow",
   303, "unable to expand memory region",
   304, "memory region size changed",
   305, "inadequate space in string region",
   306, "inadequate space in block region",

#ifdef FixedRegions
   307, "inadequate space for icode",
   308, "inadequate space for i/o buffers",
   309, "inadequate space for interpreter stack",
   310, "inadequate space in static region",
#endif					/* FixedRegions */

   311, "insufficient MAXMEM for memory expansion",

#ifdef NoCoexpr
   401, "co-expressions not implemented",
#endif					/* NoCoexpr */

   0,	0
   };

/*
 * Structures for built-in values.  Parts of some of these structures are
 *  initialized later. Since some C compilers cannot handle any partial
 *  initializations, all parts are initialized later if any have to be.
 */

/*
 * Built-in csets
 */

/*
 * &ascii; 128 bits on, second 128 bits off.
 */
struct b_cset  k_ascii = {
   T_Cset,
   128,
   cset_display(~0, ~0, ~0, ~0, ~0, ~0, ~0, ~0,
                 0,  0,  0,  0,  0,  0,  0,  0)
   };

/*
 * &cset; all 256 bits on.
 */
struct b_cset  k_cset = {
   T_Cset,
   256,
   cset_display(~0, ~0, ~0, ~0, ~0, ~0, ~0, ~0,
		~0, ~0, ~0, ~0, ~0, ~0, ~0, ~0)
   };



/*
 * Cset for &lcase; bits corresponding to lowercase letters are on.
 */
struct b_cset  k_lcase = {
   T_Cset,
   26,
#ifndef EBCDIC
   cset_display(0,  0,	0,  0,	0,  0,	~01,  03777,
		0,  0,	0,  0,	0,  0,	0,  0)
#else					/* EBCDIC */
   cset_display(0,  0,	0,  0,	0,  0,	0,  0,
		0x3fe,	0x3fe,	0x3fc,	0,  0,	0,  0,	0)
#endif					/* EBCDIC */
   };

/*
 * &ucase; bits corresponding to uppercase characters are on.
 */
struct b_cset  k_ucase = {
   T_Cset,
   26,
#ifndef EBCDIC
   cset_display(0,  0,	0,  0,	~01,  03777, 0, 0,
		0,  0,	0,  0,	0,  0,	0,  0)
#else					/* EBCDIC */
   cset_display(0,  0,	0,  0,	0,  0,	0,  0,
		0,  0,	0,  0,	0x3fe,	0x3fe,	0x3fc,	0)
#endif					/* EBCDIC */
   };

/*
 * Built-in files.
 */

struct b_file  k_errout;		/* &errout */
struct b_file  k_input;			/* &input */
struct b_file  k_output;		/* &outout */

/*
 * Keyword trapped variables.
 */

struct b_tvkywd tvky_pos; 		/* &pos */
struct b_tvkywd tvky_ran; 		/* &random */
struct b_tvkywd tvky_sub; 		/* &subject */
struct b_tvkywd tvky_trc; 		/* &trace */
struct b_tvkywd tvky_err; 		/* &error */

/*
 * Co-expression block header for &main.
 */

static struct b_coexpr *mainhead;

#if IntSize == 16
/*
 * Long integer block for &random.
 */
struct b_int long_ran = {
   T_Long,
   0L
   };
#endif					/* IntSize == 16 */

/*
 * Various constant descriptors.
 */

struct descrip blank; 			/* one-character blank string */
struct descrip emptystr; 		/* zero-length empty string */
struct descrip errout;			/* &errout */
struct descrip input;			/* &input */
struct descrip lcase;			/* string of lowercase letters */
struct descrip letr;			/* "r" */
struct descrip nulldesc;		/* null value */
struct descrip onedesc;			/* integer 1 */
struct descrip ucase;			/* string of uppercase letters */
struct descrip zerodesc;		/* integer 0 */

/*
 * init - initialize memory and prepare for Icon execution.
 */

#ifdef SCO_XENIX_LMM
extern unsigned int read();		/* not very portable ... */
#endif					/* SCO_XENIX_LMM */

init(name)
char *name;
   {
   register int i;
   word cbread, longread();
   int f;
   char c, *p;
   struct descrip *dp;
   struct header hdr;
   extern pointer malloc();

/*
 * The following code is operating-system dependent.  Make declarations
 *  that are system-dependent.
 */

#if PORT
   /* probably needs something */
#endif					/* PORT */

#if AMIGA || MSDOS
#endif					/* AMIGA || MSDOS */

#if ATARI_ST
   struct tms tp;
   extern *strchr();
#endif					/* ATARI_ST */

#if MACINTOSH
#if MPW
   extern char *brk(), *sbrk();
#endif					/* MPW */
#endif					/* MACINTOSH */

#if UNIX || VMS
   struct tms tp;
   extern char *brk(), *sbrk();
#endif					/* UNIX || VMS */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */

   extern fpetrap(), segvtrap();

   /*
    * Catch floating point traps and memory faults.
    */

/*
 * The following code is operating-system dependent. Set traps.
 */

#if PORT
   /* probably needs something */
#endif					/* PORT */

#if AMIGA
   signal(SIGFPE,fpetrap);
#endif					/* AMIGA */

#if ATARI_ST
#endif					/* ATARI_ST */

#if MACINTOSH
#if MPW
   /* This is equivalent to SIGFPE signal in the Standard Apple
      Numeric Environment (SANE) */
   {
   environment e;
   getenvironment(&e);
   setenvironment(e |
	(environment)(INVALID|UNDERFLOW|OVERFLOW|DIVBYZERO));
   sethaltvector(fpetrap);
   }
#endif					/* MPW */
#endif					/* MACINTOSH */

#if MSDOS
#if LATTICE
   signal(SIGFPE, fpetrap);
#endif					/* MICROSOFT */
#endif					/* MSDOS */

#if UNIX || VMS
#ifndef SCO_XENIX_SMM
   signal(SIGSEGV, segvtrap);
#endif					/* SCO_XENIX_SMM */
#ifdef PYRAMID
{
   struct sigvec a;

   a.sv_handler = fpetrap;
   a.sv_mask = 0;
   a.sv_onstack = 0;
   sigvec(SIGFPE, &a, 0);
   sigsetmask(1 << SIGFPE);
}
#else					/* PYRAMID */
   signal(SIGFPE, fpetrap);
#endif					/* PYRAMID */
#endif					/* UNIX || VMS */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */

#ifdef CanDump
   /*
    * If reloading from a dumped out executable, skip most of init and
    *  just set up the buffers and the timing initializations.
    */
   if (dumped)
   	goto dobufs;
#endif					/* CanDump */

   /*
    * Initializations that cannot be performed statically (at least for
    * some compilers).
    */

   k_errout.title = T_File;
   k_errout.fd = stderr;
   k_errout.status = Fs_Write;
   k_errout.fname.dword = 7;
   StrLoc(k_errout.fname) = "&errout";

   k_input.title = T_File;
   k_input.fd = stdin;
   k_input.status = Fs_Read;
   k_input.fname.dword = 6;
   StrLoc(k_input.fname) = "&input";

   k_output.title = T_File;
   k_output.fd = stdout;
   k_output.status = Fs_Write;
   k_output.fname.dword = 7;
   StrLoc(k_output.fname) = "&output";

   tvky_pos.title = T_Tvkywd;
   tvky_pos.putval = putpos;
   ((tvky_pos.kyval).dword) = D_Integer;
   IntVal(tvky_pos.kyval) = 1;
   StrLen(tvky_pos.kyname) = 4;
   StrLoc(tvky_pos.kyname) = "&pos";

   tvky_ran.title = T_Tvkywd;
   tvky_ran.putval = putran;
#if IntSize == 16
   ((tvky_ran.kyval).dword) = D_Long;
#else					/* IntSize == 16 */
   ((tvky_ran.kyval).dword) = D_Integer;
#endif					/* IntSize == 16 */
   StrLen(tvky_ran.kyname) = 7;
   StrLoc(tvky_ran.kyname) = "&random";

   tvky_sub.title = T_Tvkywd;
   tvky_sub.putval = putsub;
   StrLen(tvky_sub.kyval) = 0;
   StrLen(tvky_sub.kyname) = 8;
   StrLoc(tvky_sub.kyname) = "&subject";

   tvky_trc.title = T_Tvkywd;
   tvky_err.title = T_Tvkywd;
   tvky_trc.putval = puttrc;
   tvky_err.putval = puterr;
   ((tvky_trc.kyval).dword) = D_Integer;
   ((tvky_err.kyval).dword) = D_Integer;
   StrLen(tvky_trc.kyname) = 6;
   StrLoc(tvky_trc.kyname) = "&trace";
   StrLen(tvky_err.kyname) = 6;
   StrLoc(tvky_err.kyname) = "&error";

#if IntSize == 16
   BlkLoc(tvky_ran.kyval) = (union block *) &long_ran;
#else					/* IntSize == 16 */
   IntVal(tvky_ran.kyval) = 0;
#endif					/* IntSize == 16 */

   IntVal(tvky_trc.kyval) = 0;
   IntVal(tvky_err.kyval) = 0;
   StrLoc(tvky_sub.kyval) = "";


   StrLen(k_subject) = 0;
   StrLoc(k_subject) = "";

   StrLen(blank) = 1;
   StrLoc(blank) = " ";
   StrLen(emptystr) = 0;
   StrLoc(emptystr) = "";
   errout.dword = D_File;
   BlkLoc(errout) = (union block *) &k_errout;
   input.dword = D_File;
   BlkLoc(input) = (union block *) &k_input;
   StrLen(lcase) = 26;
   StrLoc(lcase) = "abcdefghijklmnopqrstuvwxyz";
   StrLen(letr) = 1;
   StrLoc(letr) = "r";
   nulldesc.dword = D_Null;
   IntVal(nulldesc) = 0;
   k_errorvalue = nulldesc;
   onedesc.dword = D_Integer;
   IntVal(onedesc) = 1;
   StrLen(ucase) = 26;
   StrLoc(ucase) = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
   zerodesc.dword = D_Integer;
   IntVal(zerodesc) = 0;

   maps2 = nulldesc;
   maps3 = nulldesc;
   
   /*
    * Open the icode file and read the header.
    */
   i = strlen(name);

/*
 * The following code is operating-system dependent. Open icode file.
 */

#if PORT
   f = open(name, 0);
#endif					/* PORT */

#if AMIGA
   f = open(name, O_RDONLY);
#endif					/* AMIGA */

#if ATARI_ST
   /*
    * Code to handle iconx prg  or iconx prg.icx (*note: Atari icode files
    *  must end with .icx suffix.
    */
   f = open(name,O_RDONLY | O_RAW);
   if (f < 0 && (strchr(name,'.') == NULL)) {	/* try adding .icx */
      char tname[20];
      strcat(tname,name);
      strcat(tname,".icx");
      f = open(tname,O_RDONLY | O_RAW);
      }
#endif					/* ATARI_ST */

#if MACINTOSH
#if MPW
   f = open(name, 0);
#endif					/* MPW */
#endif					/* MACINTOSH */

#if MSDOS
#if LATTICE
   f = open(name,O_RDONLY | O_RAW);
#endif					/* LATTICE */
#if MICROSOFT || TURBO
   f = open(name,O_RDONLY | O_BINARY);
#endif					/* MICROSOFT */
#endif					/* MSDOS */

#if UNIX || VMS
   f = open(name, 0);
#endif					/* UNIX || VMS */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */

   if (f < 0)
      error("can't open interpreter file");

#ifndef NoHeader
   if (lseek(f, (long)MaxHeader, 0) == -1)
      error("can't read interpreter file header");
#endif					/* NoHeader */

   if (read(f, (char *)&hdr, sizeof(hdr)) != sizeof(hdr))
      error("can't read interpreter file header");
    
   /*
    * Establish pointers to icode data regions.
    */

#ifndef FixedRegions
   code = (char *)sbrk((word)0);
#else					/* FixedRegions */
   if ((code = malloc(hdr.hsize)) == NULL)
      fatalerr(307, NULL);
#endif					/* FixedRegions */

   k_trace = hdr.trace;
   records = (word *)(uword)(code + hdr.records);
   ftab = (word *)(uword)(code + hdr.ftab);
   globals = (struct descrip *)(code + hdr.globals);
   gnames = eglobals = (struct descrip *)(code + hdr.gnames);
   statics = egnames = (struct descrip *)(code + hdr.statics);
   estatics = (struct descrip *)(code + hdr.strcons);
   strcons = (char *)estatics;

#ifdef EnvVars
   /*
    * Examine the environment and make appropriate settings.
    */
   envlook();
#endif					/* EnvVars */

   /*
    * Align bufs on a word boundary
    */

#ifdef FixedRegions
   if ((bufs = (char **)malloc(numbufs * ((BUFSIZ + sizeof(FILE **))))) ==
      NULL)
         fatalerr(308, NULL);
#else					/* FixedRegions */
   bufs = (char **)((uword)(code + hdr.hsize + 3) & ~03);
#endif					/* FixedRegions */

   bufused = (FILE **)(bufs + numbufs);

   /*
    * Convert stack sizes from words to bytes.
    */

#ifndef SCO_XENIX_LMM
   stksize *= WordSize;
   mstksize *= WordSize;
#else					/* SCO_XENIX_LMM */
   /*
    * This is a work-around for bad generated code for *= (as above)
    *  produced by the SCO XENIX C Compiler for the large memory model.
    *  It relies on the fact that WordSize is 4.
    */
   stksize += stksize;
   stksize += stksize;
   mstksize += mstksize;
   mstksize += mstksize;
#endif					/* SCO_XENIX_LMM */
 
   /*
    * Set up allocated memory.	The regions are:
    *
    *	Static memory region
    *	Allocated string region
    *	Allocate block region
    *	String qualifier list
    */

#ifdef FixedRegions
   statfree = statbase = (char *)malloc(statsize);
   statend = statfree + statsize;
   strfree = strbase = (char *)malloc(ssize);
   strend = strbase + ssize;
   blkfree = blkbase = (char *)malloc(abrsize);
   blkend = blkbase + abrsize;
   quallist = qualfree = (struct descrip **)malloc(qualsize);
   equallist = quallist + qualsize;
#else					/* FixedRegions */
   statfree = statbase = (char *)(((uword)(bufused + numbufs) + 63)  & ~077);

/*
 * The following code is operating-system dependent. Set end of static
 *  region, rounding up if necessary.
 */

#if PORT
   statend = (char *)(((uword)statbase) + mstksize + statsize);
#endif					/* PORT */

#if AMIGA
   /* uses FixedRegions */
#endif					/* AMIGA */

#if MSDOS
#ifdef LPTR
   statend =
      (char *)(((uword)statbase) + (((mstksize + statsize + 511)/512) * 512));
#else					/* LPTR */
   statend = (char *)(((uword)statbase) + mstksize + statsize);
#endif					/* LPTR */
#endif					/* MSDOS */

#if ATARI_ST || MACINTOSH || UNIX || VMS
   statend = (char *)(((uword)statbase) + mstksize + statsize);
#endif					/* ATARI_ST || MACINTOSH || UNIX ... */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */

   strfree = strbase = (char *)((uword)(statend + 63) & ~077);
   blkfree = blkbase = strend = (char *)((((uword)strbase) + ssize +
      63) & ~077);
   quallist = qualfree = equallist = (struct descrip **)(blkend =
      (char *)((((uword)(blkbase) + abrsize + 63)) & ~077));

   /*
    * Try to move the break back to the end of memory to allocate (the
    *  end of the string qualifier list) and die if the space isn't
    *  available.
    */
   if ((int)brk(equallist) == -1)
      error("insufficient memory");
   currend = (char *)sbrk(0);			/* keep track of end of memory */
#endif					/* FixedRegions */
   /*
    * Allocate stack and initialize &main.
    */

#ifndef MixedSizes
   stack = (word *)malloc(mstksize);
#else					/* MixedSizes */
   stack = (word *)malloc((int)mstksize);
#endif					/* MixedSizes */
   if (stack == NULL)
      fatalerr(309, NULL);
   mainhead = (struct b_coexpr *)stack;
   mainhead->title = T_Coexpr;
#ifndef NoCoexpr
   mainhead->es_actstk = (struct astkblk *)malloc(sizeof(struct astkblk));
   mainhead->es_actstk->nactivators = 0;
   pushact(mainhead, mainhead);
   mainhead->id = 1;
   mainhead->size = 0;
#endif					/* NoCoexpr */
   mainhead->freshblk = nulldesc;	/* &main has no refresh block. */
					/*  This really is a bug. */

   /*
    * Point &main at the co-expression block for the main procedure and set
    *  k_current, the pointer to the current co-expression, to &main.
    */
   k_main.dword = D_Coexpr;
   BlkLoc(k_main) = (union block *) mainhead;
   k_current = k_main;
   
   /*
    * Read the interpretable code and data into memory.
    */

   if ((cbread = longread(f, code, (long) hdr.hsize)) != hdr.hsize) {
      fprintf(stderr,"Tried to read %ld bytes of code, and got %ld\n",
	(long)hdr.hsize,(long)cbread);
      error("can't read interpreter code");
      }
   close(f);

/*
 * Make sure the version number of the icode matches the interpreter version.
 */

   if (strcmp((char *)hdr.config,IVersion)) {
      fprintf(stderr,"icode version mismatch\n");
      fprintf(stderr,"\ticode version: %s\n",(char *)hdr.config);
      fprintf(stderr,"\texpected version: %s\n",IVersion);
      fflush(stderr);
      if (dodump)
         abort();
      c_exit(ErrorExit);
      }

   /*
    * Resolve references from icode to runtime system.
    */
   resolve();

dobufs:
   /*
    * Mark all buffers as available.
    */
   for (i = 0; i < numbufs; i++)
      bufused[i] = NULL;

   /*
    * Buffer stdin if a buffer is available.
    */

/*
 * The following code is operating-system dependent. Buffer standard files if
 *  possible.
 */

#if PORT
   /* can't handle */
#endif					/* PORT */

#if AMIGA
   /* Level 2 I/O is automatically buffered by the system */
#endif					/* AMIGA */

#if ATARI_ST || MACINTOSH || UNIX || MSDOS
   if (numbufs >= 1) {
      setbuf(stdin, bufs[0]);
      bufused[0] = stdin;
      }
   else
      setbuf(stdin, NULL);

   /*
    * Buffer stdout if a buffer is available.
    */
   if (numbufs >= 2) {
      setbuf(stdout, bufs[1]);
      bufused[1] = stdout;
      }
   else
      setbuf(stdout, NULL);
   
   /*
    * Buffer stderr if a buffer is available.
    */
   if (numbufs >= 3 && !noerrbuf) {
      setbuf(stderr, bufs[2]);
      bufused[2] = stderr;
      }
   else
      setbuf(stderr, NULL);
#endif					/* ATARI_ST || MACINTOSH || UNIX  ... */

#if VMS
   /* can't handle */
#endif					/* VMS */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */

   /*
    * Perform any special initialization.
    */
   IconInit();

/*
 * The following code is operating-system dependent. Get start time.
 */

#if PORT
   /* needs something */
#endif					/* PORT */

#if MACINTOSH
#if MPW
   starttime = TickCount();	/* 60 ticks / second */
#endif					/* MPW */
#endif					/* MACINTOSH */

#if MSDOS || ATARI_ST || AMIGA
   time(&starttime);
#endif					/* MSDOS */

#if UNIX || VMS
   times(&tp);
   starttime = tp.tms_utime;
#endif					/* UNIX || VMS */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */

   }

/*
 * Service routines related to getting things started.
 */

/*
 * resolve - perform various fix-ups on the data read from the icode
 *  file.
 */
resolve()
   {
   register word i;
   register struct b_proc *pp;
   register struct descrip *dp;
   extern mkrec();
   extern struct b_proc *functab[];
   extern int ftsize;

   /*
    * Scan the global variable list for procedures and fill in appropriate
    *  addresses.
    */
   for (dp = globals; dp < eglobals; dp++) {
      if ((*dp).dword != D_Proc)
         continue;
      /*
       * The second word of the descriptor for procedure variables tells
       *  where the procedure is.  Negative values are used for built-in
       *  procedures and positive values are used for Icon procedures.
       */
      i = IntVal(*dp);
      if (i < 0) {
         /*
          * *dp names a built-in function, negate i and use it as an index
          *  into functab to get the location of the procedure block.
          */
         i = -i;
         if (i > ftsize - 1) {
            *dp = nulldesc;		/* undefined, set to &null */
            continue;
            }
         BlkLoc(*dp) = (union block *)functab[i-1];
         }
      else {
         /*
          * *dp names an Icon procedure or a record.  i is an offset to
          *  location of the procedure block in the code section.  Point
          *  pp at the block and replace BlkLoc(*dp).
          */
         pp = (struct b_proc *)(code + i);
         BlkLoc(*dp) = (union block *)pp;
         /*
          * Relocate the address of the name of the procedure.
          */
         StrLoc(pp->pname) = strcons + (uword)StrLoc(pp->pname);
         if (pp->ndynam == -2)
            /*
             * This procedure is a record constructor.	Make its entry point
             *	be the entry point of mkrec().
             */
            pp->entryp.ccode = mkrec;
         else {
            /*
             * This is an Icon procedure.  Relocate the entry point and
             *	the names of the parameters, locals, and static variables.
             */
            pp->entryp.icode = code + (uword)pp->entryp.icode;
            if (pp->ndynam >= 0)
               pp->filename = strcons + (uword)pp->filename;
            for (i = 0; i < pp->nparam+pp->ndynam+pp->nstatic; i++)
               StrLoc(pp->lnames[i]) = strcons + (uword)StrLoc(pp->lnames[i]);
            }
         }
      }
   /*
    * Relocate the names of the global variables.
    */
   for (dp = gnames; dp < egnames; dp++)
      StrLoc(*dp) = strcons + (uword)StrLoc(*dp);
   }

#ifdef EnvVars
/*
 * Check for environment variables that Icon uses and set system
 *  values as is appropriate.
 */
envlook()
   {
   register char *p;
   extern char *getenv();

   if ((p = getenv("TRACE")) != NULL && *p != '\0')
      k_trace = atoi(p);
   if ((p = getenv("NBUFS")) != NULL && *p != '\0')
      numbufs = atoi(p);
   if ((p = getenv("COEXPSIZE")) != NULL && *p != '\0')
      stksize = atoi(p);
   if ((p = getenv("STRSIZE")) != NULL && *p != '\0')
      ssize = atoi(p);
   if ((p = getenv("HEAPSIZE")) != NULL && *p != '\0')
      abrsize = atoi(p);
   if ((p = getenv("STATSIZE")) != NULL && *p != '\0')
      statsize = atoi(p);
   if ((p = getenv("STATINCR")) != NULL && *p != '\0')
      statincr = atoi(p);
   if ((p = getenv("MSTKSIZE")) != NULL && *p != '\0')
      mstksize = atoi(p);
#ifdef FixedRegions
   if ((p = getenv("QLSIZE")) != NULL && *p != '\0')
      qualsize = atoi(p);
#endif					/* FixedRegions */
   if ((p = getenv("ICONCORE")) != NULL && *p != '\0') {

/*
 * The following code is operating-system dependent. Check any
 *  system-dependent environment variables.
 */

#if PORT
   /* nothing to do */
#endif					/* PORT */

#if AMIGA
   if ((p = getenv("CHECKBREAK")) != NULL)
      chkbreak++;
#endif					/* AMIGA */

#if ATARI_ST || MACINTOSH || MSDOS || UNIX || VMS
   /* nothing to do */
#endif					/* ATARI_ST || MACINTOSH || MSDOS ... */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */

/*
 * The following code is operating-system dependent. Set trap to give
 *  dump on abnormal termination if ICONCORE is set.
 */

#if PORT
   /* can't handle */
#endif					/* PORT */

#if AMIGA || ATARI_ST || MACINTOSH
   /* can't handle */
#endif					/* AMIGA || ATARI_ST || MACINTOSH */

#if MSDOS
#if LATTICE
      signal(SIGFPE, SIG_DFL);
#endif					/* LATTICE */
#if TURBO
      ssignal(SIGFPE, SIG_DFL);
#endif					/* TURBO */
#endif					/* MSDOS */

#if UNIX || VMS
      signal(SIGSEGV, SIG_DFL);
#endif					/* UNIX || VMS */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */
      dodump++;
      }
   if ((p = getenv("NOERRBUF")) != NULL)
      noerrbuf++;
   }
#endif					/* EnvVars */

/*
 * Termination routines.
 */

/*
 * Produce run-time error 204 on floating point traps.
 */
#ifdef PYRAMID
fpetrap(code, subcode, sp)
int code, subcode, sp;
   {
   fatalerr(subcode == FPE_INTOVF_EXC ? 203 : 204, NULL);
   }
#else					/* PYRAMID */
fpetrap()
   {
   fatalerr(204, NULL);
   }
#endif					/* PYRAMID */

/*
 * Produce run-time error 302 on segmentation faults.
 */
segvtrap()
   {
   fatalerr(302, NULL);
   }

/*
 * error - print error message s; used only in startup code.
 */
error(s)
char *s;
   {
   fprintf(stderr, "error in startup code\n%s\n", s);
   fflush(stderr);
   if (dodump)
      abort();
   c_exit(ErrorExit);
   }

/*
 * syserr - print s as a system error.
 */
syserr(s)
char *s;
   {
   struct b_proc *bp;

   bp = (struct b_proc *)BlkLoc(argp[0]);
   if (line > 0)
      fprintf(stderr, "System error at line %ld in %s\n%s\n", (long)line,
         bp->filename, s);
   else
      fprintf(stderr, "System error in startup code\n%s\n", s);
   fflush(stderr);
   if (dodump)
      abort();
   c_exit(ErrorExit);
   }

/*
 * runerr - print message corresponding to error n and if v is nonnull,
 *  print it as the offending value.
 */

#ifdef PCIX
/*
 * For PC/IX, runerr is an assembly language routine that jumps into this
 * xruner procedure past the call to csv which occurs at the beginning of
 * all C procedures. This is necessary to defeat the stack data collision
 * testing which is done in csv in pc/ix and which would cause a loop,
 * since one of the possible reasons for calling runerr in the first place
 * might be stack/data collision.
 */
xruner(n, v)
#else					/* PCIX */
runerr(n, v)
#endif					/* PCIX */

register int n;
struct descrip *v;
   {
   register struct errtab *p;
   struct b_proc *bp;
   char *msg;

   for (p = errtab; p->errno > 0; p++)
      if (p->errno == n) {
         msg = p->errmsg;
         break;
         }

   if (line > 0) {
      if (k_error == 0) {
         bp = (struct b_proc *)BlkLoc(argp[0]);
         fprintf(stderr, "Run-time error %d\nFile %s; Line %ld\n", n,
            bp->filename, (long)line);
         }
      else {
         k_error--;
         k_errornumber = n;
         if (v!= NULL)
            k_errorvalue = *v;
         else
            k_errorvalue = nulldesc;
         k_errortext = msg;
         return;
         }
      }
   else
      fprintf(stderr, "Run-time error %d in startup code\n", n);
      fprintf(stderr, "%s\n", msg);
   if (v != NULL) {
      fprintf(stderr, "offending value: ");
      outimage(stderr, v, 0);
      putc('\n', stderr);
      }
   fflush(stderr);

#ifndef NoTraceBack
   if (line == 0 || n > 300) {	/* skip if start-up or storage problem */
      if (dodump)
         abort();
      c_exit(ErrorExit);
      }

   {
   struct pf_marker *origpfp = pfp;
   struct descrip *arg;
   struct b_proc *cproc;
   char *filename;

   fprintf(stderr, "Traceback:\n");

   /*
    * Chain back through the procedure frame markers, looking for the
    *  first one, while building a foward chain of pointers through
    *  the expression frame pointers.
    */

   for (pfp->pf_efp = NULL; pfp->pf_pfp != NULL; pfp = pfp->pf_pfp) {
      (pfp->pf_pfp)->pf_efp = (struct ef_marker *)pfp;
      }

   /* Now start from the base procedure frame marker, producing a listing
    *  of the procedure calls up through the last one.
    */

   while (pfp) {
      arg = &((struct descrip *)pfp)[-(pfp->pf_nargs) - 1];
      cproc = (struct b_proc *)BlkLoc(arg[0]);    
      filename = cproc->filename;

      xtrace(cproc, pfp->pf_nargs, &arg[0], pfp->pf_line, filename);

      /*
       * On the last call, show both the call and the offending expression.
       */
      if (pfp == origpfp) {
         ttrace(filename);
         break;
         }
 
      pfp = (struct pf_marker *)(pfp->pf_efp);
      }
   }

#endif 					/* NoTraceBack */
   if (dodump)
      abort();
   c_exit(ErrorExit);
   }

#ifndef NoTraceBack
/*
 * ttrace - show offending expression.
 */
ttrace(pfile)
char *pfile;
   {
   struct b_proc *bp;
   word nargs;
   register int n;

   fprintf(stderr, "   ");

   switch ((int)lastop) {

      case Op_Invoke:
         bp = (struct b_proc *)BlkLoc(*xargp);
         nargs = xnargs;
         if (xargp[0].dword == D_Proc)
            putstr(stderr, StrLoc(bp->pname), StrLen(bp->pname));
         else
            outimage(stderr, xargp, 0);
         putc('(', stderr);
         while (nargs--) {
            outimage(stderr, ++xargp, 0);
            if (nargs)
               putc(',', stderr);
            }
         putc(')', stderr);
         break;

      case Op_Toby:
         putc('{', stderr);
         outimage(stderr, ++xargp, 0);
         fprintf(stderr, " to ");
         outimage(stderr, ++xargp, 0);
         fprintf(stderr, " by ");
         outimage(stderr, ++xargp, 0);
         putc('}', stderr);
         break;

      case Op_Subsc:
         putc('{', stderr);
         outimage(stderr, ++xargp, 0);
         putc('[', stderr);
         outimage(stderr, ++xargp, 0);
         putc(']', stderr);
         putc('}', stderr);
         break;

      case Op_Sect:
         putc('{', stderr);
         outimage(stderr, ++xargp, 0);
         putc('[', stderr);
         outimage(stderr, ++xargp, 0);
         putc(':', stderr);
         outimage(stderr, ++xargp, 0);
         putc(']', stderr);
         putc('}', stderr);
         break;

      case Op_Bscan:
         putc('{', stderr);
         outimage(stderr, xargp, 0);
         fputs(" ? ..}", stderr);
         break;

      case Op_Coact:
         putc('{', stderr);
         outimage(stderr, ++xargp, 0);
         fprintf(stderr, " @ ");
         outimage(stderr, ++xargp, 0);
         putc('}', stderr);
         break;

      case Op_Create:
         fprintf(stderr,"{create ..}");
         break;

      case Op_Limit:
         fprintf(stderr, "limit counter: ");
         outimage(stderr, xargp, 0);
         break;
   
      default:
         bp = opblks[lastop];
         nargs = bp->nparam;
         putc('{', stderr);
         if (lastop == Op_Bang || lastop == Op_Random)
            goto oneop;
         if (bp->nparam >= 2) {
            outimage(stderr, ++xargp, 0);
            putc(' ', stderr);
            putstr(stderr, StrLoc(bp->pname), StrLen(bp->pname));
            putc(' ', stderr);
   	    }
         else
oneop:
            putstr(stderr, StrLoc(bp->pname), StrLen(bp->pname));
         outimage(stderr, ++xargp, 0);
         putc('}', stderr);
      }
	 
   if (line != 0)
      fprintf(stderr, " from line %d in %s", line, pfile);
   putc('\n', stderr);
   fflush(stderr);
   }
/*
 * xtrace - procedure *bp is being called with nargs arguments, the first
 *  of which is at arg; produce a trace message.
 */
xtrace(bp, nargs, arg, pline, pfile)
struct b_proc *bp;
word nargs;
struct descrip *arg;
int pline;
char *pfile;
   {
   register int n;

   fprintf(stderr, "   ");
   if (bp == NULL)
      fprintf(stderr, "????");

   else {
         if (arg[0].dword == D_Proc)
            putstr(stderr, StrLoc(bp->pname), StrLen(bp->pname));
         else
            outimage(stderr, arg, 0);
         arg++;
         putc('(', stderr);
         while (nargs--) {
            outimage(stderr, arg++, 0);
            if (nargs)
               putc(',', stderr);
            }
         putc(')', stderr);
      }
	 
   if (pline != 0)
      fprintf(stderr, " from line %d in %s", pline, pfile);
   putc('\n', stderr);
   fflush(stderr);
   }
#endif 					/* NoTraceBack */

/*
 * c_exit(i) - flush all buffers and exit with status i.
 */
c_exit(i)
int i;
{
   int j;

   IconTerm();		/* perform any special termination functions */
   if (tallyopt) {
      fprintf(stderr,"tallies: ");
      for (j=0; j<16; j++)
         fprintf(stderr," %ld", (long)tallybin[j]);
         fprintf(stderr,"\n");
         }
   exit(i);
}

err()
{
   syserr("call to 'err'\n");
}

/*
 * Read a long string in shorter parts. (Some systems can't handle large reads.)
 */
word longread(fd,s,len)
int fd;
char *s;
long len;
{
   long tally, n;
 
   tally = 0;
   while (len > 0) {
      n = read (fd, s, (int)((len < MaxReadCode) ? len : MaxReadCode));
      if (n <= 0)
         return tally;
      tally += n;
      s += n;
      len -= n;
      }  
   return tally;
   }

fatalerr(n, v)
int n;
struct descrip *v;
   {
   k_error = 0;
   runerr(n, v);
   }
