/*
 * Memory sizing.
 */
#ifdef FixedRegions
# undef LargeMem
# undef SmallMem
# undef IconAlloc
#else					/* FixedRegions */
# ifndef SmallMem
#  if PtrSize >= 24
#   define LargeMem
#  else 				/* PtrSize */
#   define SmallMem
#  endif				/* PtrSize */
# endif					/* SmallMem */
#endif					/* FixedRegions */

/*
 * Constants for computers with a large amount of memory.
 */

#ifdef LargeMem
#define MaxAbrSize		51200	/* size of the block region in bytes */
#define MaxStrSpace		51200	/* size of the string space in bytes */
#ifdef NoCoexpr
#define MaxStatSize		  500	/* size of static region in bytes */
#define	ActStkBlkEnts		    1   /* number of entries in an astkblk */
#else					/* NoCoexpr */
#define MaxStatSize		20480	/* size of the static region in bytes*/
#define	ActStkBlkEnts		  100   /* number of entries in an astkblk */
#endif					/* NoCoexpr */
#define MStackSize	        10000	/* size of the main stack in words */
#define StackSize		 2000	/* words in co-expression stack */
#define NumBuf			   10	/* number of i/o buffers available */
#define SSlots			   37	/* number of set slots */
#define TSlots			   37	/* number of table slots */
#endif					/* LargeMem */

/*
 * Constants for computers with a small amount of memory.
 */

#ifdef SmallMem
#define MaxAbrSize		10240	/* size of the block region in bytes */
#define MaxStrSpace		10240	/* size of the string space in bytes */
#ifdef NoCoexpr
#define MaxStatSize		  500	/* size of static region in bytes */
#define	ActStkBlkEnts		    1   /* number of entries in an astkblk */
#else					/* NoCoexpr */
#define MaxStatSize		 4096	/* size of the static region in bytes*/
#define	ActStkBlkEnts		   10   /* number of entries in an astkblk */
#endif					/* NoCoexpr */
#define MStackSize		 2000	/* size of the main stack in words */
#define StackSize		 1000	/* words in co-expression stack */
#define NumBuf			    5	/* number of i/o buffers available */
#define SSlots			   13	/* number of set slots */
#define TSlots			   13	/* number of table slots */
/*
 * MaxListSize is defined if lists are limited in size due to addressing
 * limitations of a particular architecture; it limits the size of a list-
 * element block.
 */
#define MaxListSize		 4090
#endif					/* SmallMem */

/*
 * Constants for computers with a large amount of memory but for which
 * regions cannot be expanded.
 */

#ifdef FixedRegions
#define MaxAbrSize		64000	/* size of block region in bytes */
#define MaxStrSpace		64000	/* size of string space in bytes */
#ifdef NoCoexpr
#define MaxStatSize		  500	/* size of static region in bytes */
#define	ActStkBlkEnts		    1   /* number of entries in an astkblk */
#else					/* NoCoexpr */
#define MaxStatSize		 2200	/* size of static region in bytes */
#define	ActStkBlkEnts		  100  /* number of entries in an astkblk */
#endif					/* NoCoexpr */
#define QualLstSize		 5000	/* size of qualifier pointer region */
#define MStackSize		 5000	/* size of the main stack in words */
#define StackSize		 2000	/* size of co-expression stack */
#define NumBuf			   10	/* number of i/o buffers available */
#define SSlots			   37	/* number of set slots */
#define TSlots			   37	/* number of table slots */
#endif					/* FixedRegions */

/*
 * The following code is operating-system dependent. Override
 *  memory sizing if necessary
 */

#if PORT
   /* nothing to do */
#endif					/* PORT */

#if AMIGA
#endif					/* AMIGA */

#if ATARI_ST
/*
 * Sizes needed to run reasonable icon programs on Atari 520 using ash
 */
#undef QualLstSize	
#undef MStackSize	        
#define QualLstSize		10000	/* size of qualifier pointer region */
#define MStackSize	         9500	/* size of the main stack in words */
#endif					/* ATARI_ST */

#if MACINTOSH || MSDOS || UNIX || VMS
   /* nothing to do */
#endif					/* MACINTOSH || MSDOS || UNIX || VMS */

#if VM || MVS
#endif					/* VM || MVS */

/*
 * End of operating-system specific code.
 */
