/*
 * File: rdefault.c
 *  Contents: defcset, deffile, defint, defshort, defstr
 */

#include "rt.h"

/*
 * defcset(dp,cp,buf,def) - if dp is null, default to def;
 *  otherwise, convert to cset or die trying.
 */

defcset(dp, cp, buf, def)
struct descrip *dp;
int **cp;
int *buf, *def;
   {
   if (ChkNull(*dp)) {
      *cp = def;
      return Defaulted;
      }
   if (cvcset(dp, cp, buf) == CvtFail) {
      runerr(104, dp);
      return Failure;
      }
   return Success;
   }

/*
 * deffile - if dp is null, default to def; otherwise, make sure it's a file.
 */

deffile(dp, def)
struct descrip *dp, *def;
   {
   if (ChkNull(*dp)) {
      *dp = *def;
      return Defaulted;
      }
   if (Qual(*dp) || (*dp).dword != D_File) {
      runerr(105, dp);
      return Failure;
      }
   return Success;
   }

/*
 * defint - if dp is null, default to def; otherwise, convert to integer.
 *  Note that *lp gets the value.
 */

defint(dp, lp, def)
struct descrip *dp;
long *lp;
word def;
   {
   if (ChkNull(*dp)) {
      *lp = (long)def;
      return Defaulted;
      }
   if (cvint(dp, lp) == CvtFail) {
      runerr(101, dp);
      return Failure;
      }
   return Success;
   }

/*
 * defshort - if dp is null, default to def; otherwise, convert to short
 *  integer.  The result is an integer value in *dp.
 */

defshort(dp, def)
struct descrip *dp;
int def;
   {
   long l;

   if (ChkNull(*dp)) {
      MkInt((int)def, dp);
      return Defaulted;
      }
   switch (cvint(dp, &l)) {

      case T_Integer:
         MkInt(l, dp);
	 return Success;

      case T_Long:
         runerr(205, dp);
         return Failure;

      default:
	 runerr(101, dp);
         return Failure;
      }
   }

/*
 * defstr - if dp is null, default to def; otherwise, convert to string.
 *  *dp gets a descriptor for the resulting string.  buf is used as
 *  a scratch buffer for the conversion (if necessary).
 */

defstr(dp, buf, def)
struct descrip *dp;
char *buf;
struct descrip *def;
   {
   if (ChkNull(*dp)) {
      *dp = *def;
      return Defaulted;
      }

   if (cvstr(dp, buf) == CvtFail) {
      runerr(103, dp);
      return Failure;
      }

   return Success;
   }
