/*
 * This routine switches contexts for coexpressions in the SAS/C
 *  version of Icon.  Each coexpression is implemented by a SAS/C
 *  coprocess.  The main coprocess is a dispatcher process, which
 *  cocalls whatever coexpression is to run next.  The coexpression
 *  does a coreturn to request the dispatcher to switch to another
 *  coexpression, or to terminate a coexpression which has been
 *  garbage collected.
 */

#include "config.h"
#include "rt.h"
#include <coproc.h>

static coproc_t main,           /* id for main Icon coprocess */
                quitter = 0;    /* id of coprocess killing another */

/*
 * Each coprocess starts its life executing coentry.  coentry merely
 *  calls the interpreter to start things off.
 */
void *coentry(void *arg)        /* argument not used */
   {
      quiet(1);                 /* turn off library diagnostics    */
                                /* for this process                */
      interp(0, 0);
      if (coproc(SELF) != main)
         syserr("interp() returned in coswitch");
      return NULL;
   }

/*
 * This is the coprocess dispatcher.
 */
void codisp()
   {
      coproc_t next;            /* the next coproc to run */

      ((struct b_coexpr *)BlkLoc(k_main))->cstate [1]  =
         main = costart(&coentry, NULL);
                                        /* start Icon main coprocess */
      next = (coproc_t) cocall(main, NULL);     /* and run it  */
      while (next != NULL) {            /* return data is the next   */
                                        /* one to run                */
         if (quitter) {                 /* if next is to be killed   */
            cocall(next, (void *) -1);  /* tell it to die            */
            next = (coproc_t) cocall(quitter, NULL);
                                        /* then go back to killer    */
         }
         else next = (coproc_t) cocall(next, NULL);
                                        /* run the next coexpression */
      }
   }

/*
 * coswitch is called to switch contexts.  The first flag indicates
 *  whether this is the first call or a resumption; a first of -1
 *  is used to terminate the 'new' coexpression when it is garbage
 *  collected.
 */
void coswitch(old_cs, new_cs, first)
   int * old_cs, * new_cs;
   int first;
   {
      coproc_t old, new;
      void *Iquit;

      old = old_cs[1];                  /* get old and new coproc ids*/
      new = new_cs[1];
      if (first == -1)                  /* if new is to die          */
         {
            if (new == 0) return;       /* maybe coexpr never ran    */
            quitter = old;              /* note who to resume after  */
                                        /* new process dies          */
            coreturn((void *)new);      /* tell dispatcher to finish */
                                        /* new                       */
            quitter = 0;
            return;
         }

      if (first == 0)                   /* if new never ran before   */
         new = new_cs[1] = costart(&coentry, NULL);
                                        /* create coproc for it      */
      Iquit = coreturn((void *) new);   /* tell dispatcher to run new*/
      if (Iquit != NULL)                /* see if we've been killed  */
         coexit(NULL);                  /* if so, die agreeably      */
      return;                           /* else return to run again  */
   }
