         /* A version of EXTCALL for using ISPF services from
            Icon.  See ISPF.ICN for intended usage. */

#include "config.h"
#include "rt.h"
#include "rproto.h"

#ifdef ExternalFunctions

static struct descrip spfretv;
static char spfup;      /* is ISPF around? */
static char first;

static dptr extqry(), extexec(), extrepl(), extcopy();
static int makemsg();

extern __asm ispexec();
extern __asm isplink();
extern __asm ispqry();

#define compare(a,b) (StrLen(a) != sizeof(b) - 1 || \
                      memcmp(StrLoc(a), (b), sizeof(b) - 1))
               /* efficient compare of string to constant */

/*
 * extcall - stub procedure for external call interface.
 */
dptr extcall(dargv, argc, ip)
dptr dargv;
int argc;
int *ip;
   {
   char sbuf[MaxCvtLen];

   if (cvstr(dargv, sbuf) == CvtFail) {
      *ip = 103;
      return dargv;
   }
   if (!compare(*dargv, "ispexec")) return extexec(dargv, argc, ip);
   if (!compare(*dargv, "ispqry")) return extqry(dargv, argc, ip);
   if (!compare(*dargv, "ispcopy")) return extcopy(dargv, argc, ip);
   if (!compare(*dargv, "isprepl")) return extrepl(dargv, argc, ip);
   *ip = 216;                   /* no external function to find */
   return (dptr)NULL;
   }

static dptr extqry(dargv, argc, ip)  /* see if ISPF is around */
dptr dargv;
int argc;
int *ip;
   {
   if (!spfup) spfup = ispqry() ? 'N' : 'Y';
   if (spfup == 'N') return NULL;    /* fail if nowhere in sight */
   return &nulldesc;                 /* else return &null */
   }

static dptr extexec(dargv, argc, ip) /* issue an ISPEXEC request */
dptr dargv;
int argc;
int *ip;
   {
   char sbuf[MaxCvtLen];
   int spfrc;

   if (!spfup) spfup = ispqry() ? 'N' : 'Y';
   if (argc < 2) {      /* too few arguments */
      *ip = 103;
      return NULL;
   }
   if (spfup == 'N') {
      if (makemsg("ISPF not active", ip))
         return &spfretv;
      else return (dptr) NULL;
   }
         /* Usually, we would like Icon to get control back after
            an error.  So ask ISPF to let us continue.  The applic-
            ation can override this if it wishes. */
   if (first == 0) {
      first = 1;
      ispexec(@21,"CONTROL ERRORS RETURN");
   }
   ++dargv;
   if (cvstr(dargv, sbuf) == CvtFail) {
      *ip = 103;     /* argument not string */
      return dargv;
   }
   spfrc = ispexec(@StrLen(*dargv), StrLoc(*dargv));
         /* There must be a blank after the command for ISPEXEC
            to work correctly.  We depend on our caller to put
            it there. */
   if (spfrc < 12) {    /* nothing terrible happened */
      MakeInt(spfrc, &spfretv);      /* return integer return code */
      return &spfretv;
   } else {             /* bad news, ask ISPF for message */
      int spflen = 78;
      char spfmsg[79];
      if (isplink("VCOPY   ", "ZERRLM ", &spflen, spfmsg, "MOVE ")) {
         if (makemsg("ISPF error of unknown type", ip))
            return &spfretv;
         else return (dptr) NULL;
      }
      spfmsg[78] = 0;
      if (makemsg(spfmsg, ip))
         return &spfretv;      /* after error, return message text */
      else return (dptr) NULL;
   }
   }

static dptr extrepl(dargv, argc, ip) /* replace an ISPF variable */
dptr dargv;
int argc;
int *ip;
   {
   char sbuf1[MaxCvtLen];
   char sbuf2[MaxCvtLen];
   int spfrc;

   if (!spfup) spfup = ispqry() ? 'N' : 'Y';
   if (argc < 3) {      /* too few arguments */
      *ip = 103;
      return NULL;
   }
   if (spfup == 'N') {
      if (makemsg("ISPF not active", ip))
         return &spfretv;
      else return (dptr) NULL;
   }
   if (first == 0) {
      first = 1;
      ispexec(21,"CONTROL ERRORS RETURN");
   }
   ++dargv;
   if (cvstr(dargv, sbuf1) == CvtFail) {
      *ip = 103;     /* argument not string */
      return dargv;
   }
   if (cvstr(dargv+1, sbuf2) == CvtFail) {
      *ip = 103;
      return dargv+1;
   }
   spfrc = isplink("VREPLACE ", StrLoc(*dargv), @StrLen(dargv[1]),
                   StrLoc(dargv[1]));   /* replace the variable */
               /* caller must terminate variable name with blank */
   if (spfrc < 12) {    /* no serious error...return integer code */
      MakeInt(spfrc, &spfretv);
      return &spfretv;
   } else {             /* bad news, ask ISPF for message */
      int spflen = 78;
      char spfmsg[79];
      if (isplink("VCOPY   ", "ZERRLM ", &spflen, spfmsg, "MOVE ")) {
         if (makemsg("ISPF error of unknown type", ip))
            return &spfretv;
         else return (dptr) NULL;
      }
      spfmsg[78] = 0;
      if (makemsg(spfmsg, ip))
         return &spfretv;   /* after error, return message text */
      else return (dptr) NULL;
   }
   }

static dptr extcopy(dargv, argc, ip) /* copy an ISPF variable's value */
dptr dargv;
int argc;
int *ip;
   {
   char sbuf[MaxCvtLen];
   int spflen;
   char *spfdata;
   int spfrc;

   if (!spfup) spfup = ispqry() ? 'N' : 'Y';
   if (argc < 2) {
      *ip = 103;     /* too few arguments */
      return NULL;
   }
   if (spfup == 'N') {
      if (makemsg("!ISPF not active", ip))
         return &spfretv;
      else return (dptr) NULL;
   }
   if (first == 0) {
      first = 1;
      ispexec(21,"CONTROL ERRORS RETURN");
   }
   ++dargv;
   if (cvstr(dargv, sbuf) == CvtFail) {
      *ip = 103;     /* argument not string */
      return dargv;
   }
   spfrc = isplink("VCOPY   ", StrLoc(*dargv), &spflen, &spfdata,
                   "LOCATE ");
               /* caller must terminate variable name with blank */
   if (spfrc == 8) return NULL;      /* fail if var not found */
   else if (spfrc == 0) {
      if (strreq(spflen+1) == Error) {
         *ip = 0;
         return 0;
      }
      StrLoc(spfretv) = alcstr("=", 1);   /* flag return as value,
                                             not message */
      (void) alcstr(spfdata, spflen);
      StrLen(spfretv) = spflen+1;
      return &spfretv;
   } else {    /* ugly...return ISPF error message, prefixed by ! to
                  distinguish it from a variable value */
      int spflen = 78;
      char spfmsg[80];
      if (isplink("VCOPY   ", "ZERRLM ", &spflen, spfmsg+1, "MOVE ")) {
         if (makemsg("!ISPF error of unknown type", ip))
            return &spfretv;
         else return (dptr) NULL;
      }
      spfmsg[0] = '!';
      spfmsg[79] = 0;
      if (makemsg(spfmsg, ip))
         return &spfretv;
      else return (dptr) NULL;
   }
   }

static int makemsg(text, ip)    /* turn a message into an Icon string */
char *text;
int *ip;
   {
   int len;

   len = strlen(text);
   if (strreq(len) == Error) {
      *ip = 0;
      return 0;
   }
   StrLoc(spfretv) = alcstr(text, len);
   StrLen(spfretv) = len;
   return 1;
   }
#else                                   /* ExternalFunctions */
static char x;                  /* prevent empty module */
#endif                                  /* ExternalFunctions */
