/*
 * File: fxwicon.r
 *  Contents: XIconic, XIconImage, XIconLabel, XMoveIcon, XWindowLabel
 */

#ifdef XIcon

#define RootState IconicState+1


"XIconic(w,s) - iconify/de-iconify a window or return state"

function{0,1} XIconic(argv[argc])
   abstract {
      return string
   }
   body {
      wbp    _w_;
      int warg = 0;
      tended char *tmp;

      OptWindow;

      {
      STDLOCALS(_w_);

#ifndef PresentationManager
      if (ws->iconic == RootState)           /* can't (de)iconify the root */
         fail;
#else                                   /* PresentationManager */
      /* doesn't apply to bitmaps */
      if (!stdwin) fail;
#endif                                  /* PresentationManager */

      if (warg < argc) {
         if (!cnv:C_string(argv[warg],tmp))        /* convert to check arg */
            runerr(103,argv[warg]);
         if (!strcmp(tmp, "icon")) {         /* request to iconify */
#ifndef PresentationManager
            if (ws->iconic == NormalState)
#ifdef Iconify
               XIconifyWindow(stddpy, stdwin, wd->screen);
#else                                   /* Iconify */
               fail;
#endif                                  /* Iconify */
#else                                   /* PresentationManager */
            if (ISICONIC(_w_)) fail;
            SETICONIC(_w_);
            WinShowWindow(ws->hwndFrame, FALSE);
#endif                                  /* PresentationManager */
            }
         else if (!strcmp(tmp, "window")) {  /* request to de-iconify */
#ifndef PresentationManager
            if (ws->iconic == IconicState)
               XMapWindow(stddpy, stdwin);
#else                                   /* PresentationManager */
            int height;

            /* if we are already windowed, fail */
            if (!ISICONIC(_w_)) fail;
            CLRICONIC(_w_);
            height = ws->height + TITLEHEIGHT + (BORDERHEIGHT << 1);
            WinSetWindowPos(ws->hwndFrame, HWND_TOP,
                            ws->posx,
                            ScreenHeight - ws->posy - height,
                            ws->width + (BORDERWIDTH << 1),
                            height,
                            SWP_SIZE | SWP_SHOW | SWP_MOVE |
                            SWP_ACTIVATE | SWP_ZORDER);
#endif                                  /* PresentationManager */
            }
         else                                /* invalid argument */
            fail;
         }
#ifndef PresentationManager
      switch (ws->iconic) {
      case RootState:
         return C_string "root";
      case NormalState:
         return C_string "window";
      case IconicState:
         return C_string "icon";
         }
#else                                   /* PresentationManager */
      return C_string ((ISICONIC(_w_)) ? "icon" : "window");
#endif                                  /* PresentationManager */
      }
      }
end


#ifndef PresentationManager     /* avoiding inclusion */
/*
 * remap a window ... this time with an iconwin
 */
int remap(w,x,y)
wbp w;
int x,y;
   {
   XSizeHints size_hints;
   XWMHints   *wmhints;
   Window     root;
   unsigned int width, height, bw, depth;
   STDLOCALS(w);

   XGetSizeHints(stddpy, stdwin, &size_hints, XA_WM_NORMAL_HINTS);
   wmhints = XGetWMHints(stddpy, stdwin);
   if (ws->iconwin)
      XDestroyWindow(stddpy, ws->iconwin);
   if (stdwin)
      XDestroyWindow(stddpy, stdwin);

   ws->win = XCreateSimpleWindow(stddpy, DefaultRootWindow(stddpy),
                                 ws->posx, ws->posy, ws->width,
                                 ws->height, 4, wc->fg->c, wc->bg->c);
   XSetStandardProperties(stddpy, ws->win, ws->windowlabel,
                          ws->iconlabel, 0, 0, 0, &size_hints);
   XSelectInput(stddpy, ws->win, ExposureMask | KeyPressMask |
                ButtonPressMask | ButtonReleaseMask | ButtonMotionMask |
                StructureNotifyMask);

   ws->iconwin = XCreateSimpleWindow(stddpy, DefaultRootWindow(stddpy),
                                     ws->iconx, ws->icony, ws->iconw,
                                     ws->iconh, 2, wc->fg->c, wc->bg->c);
   XSelectInput(stddpy, ws->iconwin,
                ExposureMask | KeyPressMask | ButtonPressMask);

   wmhints->flags |= IconPositionHint;
   wmhints->icon_x = x;
   wmhints->icon_y = y;
   wmhints->initial_state = ws->iconic;
   wmhints->icon_window = ws->iconwin;
   wmhints->flags |= IconWindowHint;
   XSetWMHints(stddpy, ws->win, wmhints);
   CLREXPOSED(w);
   XMapWindow(stddpy, ws->win);
   if (ws->iconic == NormalState) {
      while (!ISEXPOSED(w))
         if (pollevent() == -1) return -1;
      }
   ws->iconx = x;
   ws->icony = y;
   XSync(stddpy, False);
   XFree((char *)wmhints);
   return 1;
   }


"XIconImage(w,x) - set w's icon pixmap"

function{0,1} XIconImage(argv[argc])
   abstract {
      return string
   }
   body {
      wbp    _w_;
      int warg = 0;
      tended char *tmp;
      int    len;
      Pixmap pix;
      int status;

      OptWindow;

      {
      STDLOCALS(_w_);

      if (ws->iconic == RootState)
         fail;
      if (warg < argc) {
         /* get the preloaded pixmap image */
         if (is:file(argv[warg]) &&
             (BlkLoc(argv[warg])->file.status & Fs_Window)) {
            wbp _x_;
            strcpy(ws->iconimage, "");
            _x_ = (wbp)BlkLoc(argv[warg])->file.fd;
            pix = XCreatePixmap(stddpy, DefaultRootWindow(stddpy),
                                _x_->window->width, _x_->window->height,
                                DefaultDepth(stddpy,wd->screen));

            XCopyArea(stddpy, _x_->window->pix, pix, wd->icongc, 0, 0,
                      _x_->window->width, _x_->window->height, 0, 0);
            if (ws->iconpix) {
               XSync(stddpy, False);
               XFreePixmap(stddpy, ws->iconpix);
               }
            ws->iconpix = pix;
            ws->iconw = _x_->window->width;
            ws->iconh = _x_->window->height;
            if (!ws->iconx && !ws->icony) {
               ws->iconx = ws->x;
               ws->icony = ws->y;
               }
            if (remap(_w_,ws->iconx,ws->icony) == -1) runerr(217);
            }
         /* get the pixmap file named by x */
         else if (is:string(argv[warg])) {
            int height, width;
            if (!cnv:C_string(argv[warg],tmp))
               runerr(103,argv[warg]);
            strcpy(ws->iconimage, tmp); 
            if (ws->iconwin == (Window) NULL) makeIcon(_w_,0,0);
            else {
               pix = loadimage(_w_,ws->iconimage, &height, &width, 0, &status);
               if (pix == (Pixmap) NULL)
                  fail;
               XCopyArea(stddpy, pix, ws->iconwin, wd->icongc,
                         0, 0, width, height, 0, 0);
               if (ws->iconpix) {
                  XSync(stddpy, False);
                  XFreePixmap(stddpy, ws->iconpix);
                  }
               ws->iconpix = pix;
               }
            }
         else
            fail;
         }
      len = strlen(ws->iconimage);
      Protect(tmp = alcstr(ws->iconimage, len), runerr(0));
      return string(len, tmp);
      }
      }
end


"XIconLabel(w,s) - set w's icon label to string s"

function{1} XIconLabel(argv[argc])
   abstract {
      return string
   }
   body {
      wbp    _w_;
      int warg = 0;
      tended char *tmp;
      int    len;

      OptWindow;

      {
      STDLOCALS(_w_);

      if (ws->iconic == RootState)
         fail;
      if (warg < argc) {
         if (!cnv:C_string(argv[warg],tmp))         /* convert to check arg */
               runerr(103,argv[warg]);
            strcpy(ws->iconlabel, tmp);
            XSetIconName(stddpy, stdwin, tmp);
            if (ws->iconic == IconicState && !ws->iconpix) {
               XClearWindow(stddpy, ws->iconwin);
               XDrawString(stddpy, ws->iconwin, wd->icongc, 4,
                           wd->fonts->fsp->max_bounds.ascent + 2,
                           ws->iconlabel, strlen(ws->iconlabel));
            }
         }
         len = strlen(ws->iconlabel);
         Protect(tmp = alcstr(ws->iconlabel, len), runerr(0));
         return string(len, tmp);
      }
   }
end


"XMoveIcon(w,x,y) - move icon to x,y"

function{1} XMoveIcon(argv[argc])
   declare {
      int _x_, _y_;
   }
   abstract {
      return file
   }

   body {
      wbp _w_;
      int warg = 0;

      OptWindow;

      if (argc - warg == 0) _x_ = 0;
      else if (!cnv:C_integer(argv[warg], _x_)) runerr(101,argv[warg]);
      if (argc - warg < 2) _y_ = 0;
      else if (!cnv:C_integer(argv[warg+1], _y_)) runerr(101,argv[warg+1]);

      if (_w_->window->iconic == RootState)
         fail;

      if (_w_->window->iconwin == (Window) NULL)
         makeIcon(_w_,_x_,_y_);
      if (remap(_w_, _x_, _y_) == -1) runerr(217);
      ReturnWindow;
      }
end
#endif                                  /* Avoid inclusion for now XXX */


"XWindowLabel(w,s) - set w's window label to string s"

function{1} XWindowLabel(argv[argc])
   abstract {
      return string
   }
   body {
      wbp    _w_;
      int    warg = 0;
      tended char *tmp;
      int    len;
#ifdef PresentationManager
      char buf[MAXLABEL];
#endif                                  /* PresentationManager */

      OptWindow;

      {
      STDLOCALS(_w_);

      if (warg < argc) {
#if CSET2
         if (!cnv:C_string(r_args[1 + warg],tmp))       /* convert to check arg */
            runerr(103,r_args[1 + warg]);
#else                                   /* CSET2 - BUG!!! */
         if (!cnv:C_string(argv[warg],tmp))       /* convert to check arg */
            runerr(103,argv[warg]);
#endif                                  /* CSET2 - BUG!!! */
#ifndef PresentationManager
         mystrncpy(ws->windowlabel, tmp, MAXLABEL-1);
         XStoreName(stddpy, stdwin, tmp);
         }
      len = strlen(ws->windowlabel);
      Protect(tmp = alcstr(ws->windowlabel, len), runerr(0));
#else                                   /* PresentationManager */
        mystrncpy(buf, tmp, MAXLABEL - 1);
        if (stdwin)
          WinSetWindowText(ws->hwndFrame, buf);
        }
      else
        WinQueryWindowText(ws->hwndFrame, 256, buf);
      len = strlen(buf);
      Protect(tmp = alcstr(buf, len), runerr(0));
#endif                                  /* PresentationManager */
      return string(len, tmp);
      }
      }
end
#else                                   /* XIcon */
static char x;                  /* avoid empty module */
#endif                                  /* XIcon */
